---
---

@import "{{ site.theme }}";

/* Pad image bottom */
.wrapper img {
    padding-bottom: 1%;
}

/* Remove bullets from ordered lists */
ul {
  list-style-type: none;
  padding: 0;
  margin: 0;
}

/* Make content wider */
.wrapper {
	width: 1200px;
}

section {
	width: 860px;
}

footer {
  width: 280px;
}
